package com.sf.vertx.init;

import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;

import com.sf.vertx.api.pojo.AppConfig;
import com.sf.vertx.api.pojo.GatewayInterface;
import com.sf.vertx.api.pojo.Node;
import com.sf.vertx.api.pojo.SacService;
import com.sf.vertx.api.pojo.VertxConfig;
import com.sf.vertx.arithmetic.roundRobin.SacLoadBalancing;
import com.sf.vertx.arithmetic.roundRobin.WeightedRoundRobin;
import com.sf.vertx.constans.RedisConfig;
import com.sf.vertx.handle.BodyHandler;
import com.sf.vertx.handle.ProxyHandler;
import com.sf.vertx.security.MainSecurity;
import com.sf.vertx.service.AppConfigService;
import com.sf.vertx.service.impl.AppConfigServiceImpl;

import io.vertx.core.Future;
import io.vertx.core.Vertx;
import io.vertx.core.VertxOptions;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpServer;
import io.vertx.core.http.HttpServerOptions;
import io.vertx.core.http.HttpServerRequest;
import io.vertx.core.net.SocketAddress;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.client.WebClient;
import io.vertx.httpproxy.HttpProxy;
import io.vertx.httpproxy.ProxyContext;
import io.vertx.httpproxy.ProxyInterceptor;
import io.vertx.httpproxy.ProxyResponse;
import lombok.extern.slf4j.Slf4j;

/***
 * 动态构建vertx服务
 * 
 * @author xy
 *
 */
@Slf4j
@Order(value = 10)
@Component
public class DynamicBuildServer implements ApplicationRunner {
	public static final String SAC_APP_HEADER_KEY = "sacAppCode";
	private static ConcurrentHashMap<Integer, SacLoadBalancing> SAC_LOADBALANCING_MAP = new ConcurrentHashMap<Integer, SacLoadBalancing>();
	@Value("${server.vertx.server.default.port}")
	private Integer serverDefaultPort;

	@Autowired
	private AppConfigService appConfigService;

	@Autowired
	private RedisConfig redisConfig;

	@Override
	public void run(ApplicationArguments args) throws Exception {
		// 初始化redis key
		redisConfig.init();
		// 加载vertx、应用配置
		appStartLoadData();
	}

	/***
	 * 应用启动, 从redis读取配置,初始化vertx服务
	 */
	private void appStartLoadData() {
		// TODO 编解码线程池,后面优化协程等方式
		Vertx VERTX = Vertx.vertx(new VertxOptions().setWorkerPoolSize(20));
		// 创建HTTP监听
		// 所有ip都能访问
		HttpServerOptions httpServerOptions = new HttpServerOptions().setHost("0.0.0.0");
		HttpServer server = VERTX.createHttpServer(httpServerOptions);
		Router mainHttpRouter = Router.router(VERTX);
		VertxConfig vertxConfig = appConfigService.loadVertxConfig();
		Integer serverPort = (vertxConfig == null || vertxConfig.getPort() == null) ? serverDefaultPort
				: vertxConfig.getPort();
		log.info("serverPort:{}", serverPort);
		server.requestHandler(mainHttpRouter).listen(serverPort, h -> {
			if (h.succeeded()) {
				log.info("HTTP端口监听成功:{}", serverPort);
			} else {
				log.error("HTTP端口监听失败:{}", serverPort);
			}
		});

		ConcurrentHashMap<String, AppConfig> cacheAppConfig = appConfigService.loadAllConfig();
//		HttpClientOptions clientOptions = new HttpClientOptions();
//		clientOptions.setMaxPoolSize(20); // 最大连接池大小
//		clientOptions.setConnectTimeout(5000); // 连接超时 毫秒
//		clientOptions.setHttp2KeepAliveTimeout(1);
//		clientOptions.setIdleTimeout(1000); // 连接空闲超时 毫秒
//		HttpClient proxyClient = VERTX.createHttpClient(clientOptions);
		HttpClient proxyClient = VERTX.createHttpClient();
		HttpProxy proxy = HttpProxy.reverseProxy(proxyClient);
		proxy.originSelector(request -> Future.succeededFuture(resolveOriginAddress(cacheAppConfig, request)));
		proxy.addInterceptor(new ProxyInterceptor() {
			@Override
			public Future<ProxyResponse> handleProxyRequest(ProxyContext context) {
				// 修改uri context.request().setURI();
				String sacAppHeaderKey = context.request().headers().get(DynamicBuildServer.SAC_APP_HEADER_KEY);
				log.info("addInterceptor uri appCode:{}", sacAppHeaderKey);
				// 判断是否需要加解析
				if (AppConfigServiceImpl.appDataSecurity(sacAppHeaderKey)) {
					
					//String data = decode(null, sacAppHeaderKey);
							
				}
				return context.sendRequest();
			}
		});
		// mainHttpRouter.route().handler(ProxyHandler.create(proxy));
		WebClient mainWebClient = WebClient.create(VERTX);
		mainHttpRouter.route().handler(BodyHandler.create()).handler(ProxyHandler.create(mainWebClient,proxy));
	}

	public SocketAddress resolveOriginAddress(ConcurrentHashMap<String, AppConfig> cacheAppConfig,
			HttpServerRequest request) {
		String appCode = request.getHeader(SAC_APP_HEADER_KEY);
		log.info("uri appCode:{}", appCode);
		// TODO 不存在, 抛异常给前端

		AppConfig appConfig = cacheAppConfig.get(appCode);
		if (appConfig != null) {
			Integer environmentId = null;
			SacLoadBalancing sacLoadBalancing = null;
			// 2、是否存在server服务配置
			if (appConfig.getService() != null && appConfig.getService().size() > 0) {
				for (SacService service : appConfig.getService()) {
					// uri是否匹配
					boolean match = false;
					for (GatewayInterface gatewayInterface : service.getUriList()) {
						String domain = request.authority().port() == -1 ? request.authority().host()
								: request.authority().host() + ":" + request.authority().port();
						// uri匹配(正则或全量匹配)
						match = gatewayInterface.isUriRegular() ? regexMatch(gatewayInterface.getUri(), request.uri())
								: StringUtils.equals(gatewayInterface.getUri(), request.uri());
						match = match ? StringUtils.equals(gatewayInterface.getMethod(), request.method().name())
								: false;
						// domain匹配
						if (match) {
							boolean domainVertify = service.getRouter() != null
									&& StringUtils.isNotBlank(service.getRouter().getDomain())
									&& StringUtils.equals(service.getRouter().getDomain(), domain) ? true : false;
							if (domainVertify) {
								environmentId = service.getRouter().getEnvironmentId();
							} else {
								environmentId = appConfig.getEnvironmentConfig().getDefaultId();
							}
							break;
						}
					}
				}
			} else {
				environmentId = appConfig.getEnvironmentConfig().getDefaultId();
			}
			// 初始化负载均衡
			if (SAC_LOADBALANCING_MAP.get(environmentId) != null) {
				sacLoadBalancing = SAC_LOADBALANCING_MAP.get(environmentId);
			} else {
				// 并发影响不大, 只要初始化成功一次即可
				List<Node> nodeList = appConfig.getEnvironmentConfig().getEnvironmentGroup().get(environmentId);
				if (nodeList != null && nodeList.size() > 0) {
					// 初始化负载均衡算法
					sacLoadBalancing = roundRobin(nodeList);
					SAC_LOADBALANCING_MAP.put(environmentId, sacLoadBalancing);
				} else {
					// TODO 抛出异常
				}
			}

			// TODO 区分https、http
			Node node = sacLoadBalancing.selectNode();
			SocketAddress socketAddress = SocketAddress.inetSocketAddress(node.getPort(), node.getIp());
			log.info("负载均衡跳转地址:{}", socketAddress.host() + ";" + socketAddress.port());
			return socketAddress;
		}
		// TODO 如果没找到, 如何处理
		return null;
	}

//	private String getUriAppCode(String uri) {
//		// 1、判断appCode
//		int count = 0;
//		int appCodeLen = 16;
//		StringBuffer appCode = new StringBuffer();
//		for (int i = 0; i < uri.length(); i++) {
//			// 限制长度
//			if (appCode.length() == appCodeLen) {
//				break;
//			}
//			char ch = uri.charAt(i);
//			if (ch == '/') {
//				if (++count == 2) {
//					break;
//				}
//			} else {
//				appCode.append(ch);
//			}
//		}
//		return appCode.toString();
//	}

	private static boolean regexMatch(String pattern, String target) {
		return Pattern.matches(pattern, target);
	}

	/***
	 * 解密
	 */
	private static String decode(AppConfig appConfig, String bodyJson) {
		String algorithm = appConfig.getDataSecurity().getAlgorithm();
		String data = null;
		switch (algorithm) {
		case "aes":
			// 解密
			String key = appConfig.getDataSecurity().getKey();
			data = MainSecurity.aesDecrypt(bodyJson, key);
			break;
		default:
			break;
		}
		return data;
	}

	/***
	 * 加密
	 */
	private static String encryption(AppConfig appConfig, String responseData) {
		String algorithm = appConfig.getDataSecurity().getAlgorithm();
		String data = null;
		switch (algorithm) {
		case "aes":
			String key = appConfig.getDataSecurity().getKey();
			data = MainSecurity.aesEncrypt(responseData.toString(), key);
			break;
		default:
			break;
		}
		return data;
	}

	private static SacLoadBalancing roundRobin(List<Node> nodeList) {
		WeightedRoundRobin weightedRoundRobin = new WeightedRoundRobin();
		for (Node node : nodeList) {
			int weight = node.getWeight() != null ? node.getWeight() : 1;
			node.setWeight(weight);
			node.setCurrentWeight(weight);
			node.setEffectiveWeight(weight);
			WeightedRoundRobin.totalWeight += node.getEffectiveWeight();
		}
		weightedRoundRobin.init(nodeList);
		return weightedRoundRobin;
	}
}
